import math

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from chester.run.user_classes import TimeSeriesHandler
from chester.util import remove_prefix_suffix
from chester.zero_break.problem_specification import DataInfo


class TimeSeriesFeatureStatistics:
    def __init__(self,
                 data_info: DataInfo,
                 time_series_handler: TimeSeriesHandler,
                 max_print=None,
                 ts_cols=None):
        self.data_info = data_info
        self.max_print = max_print
        self.ts_cols = ts_cols
        self.time_series_handler = time_series_handler

        self.time_cols = self.data_info.feature_types_val["time"]
        self.data = self.data_info.data[list(set(self.ts_cols + self.time_cols))]
        # if self.data.columns.duplicated().any():
        #     self.data = self.data.loc[:, ~ self.data.columns.duplicated()]
        self.data = self.data.sample(min(10000, len(self.data)))
        self.time_frequency = time_series_handler.time_frequency

    @staticmethod
    def discover_scale(column):
        median = column.median()
        if median >= 31536000:
            return "years", 31536000
        elif median > 2592000:
            return "months", 2592000
        elif median > 604800:
            return "weeks", 604800
        elif median > 86400:
            return "days", 86400
        elif median >= 3600:
            return "hours", 3600
        elif median >= 60:
            return "minutes", 60
        else:
            return "seconds", 1

    def plot_time_between_events(self, time_col, plot=False):
        # all_columns = self.data_info.get_all_features()
        all_columns = [name for name in self.data_info.data.columns if name.startswith("ts_")]

        _, metric_options = self.time_series_handler.lag_values
        freq_cols = [name for name in all_columns if
                     name.startswith("ts_freq_") and name.endswith(time_col) and metric_options[0] in name]
        freq_cols.sort()
        dim_plots = math.floor(math.sqrt(len(freq_cols)))  # plot dim*dim graphs, on the same plot
        if dim_plots == 0:
            plt.close()
            return None
        # Create a subplot grid for the histograms
        fig, axs = plt.subplots(dim_plots, dim_plots, figsize=(10, 7))

        # Plot a histogram for each column in freq_cols and add a title
        for i, col in enumerate(freq_cols):
            if i > dim_plots * dim_plots - 1:
                break
            row_index = i // dim_plots
            col_index = i % dim_plots

            clean_col = remove_prefix_suffix(string=col, prefix="ts_freq_", suffix=f"_{time_col}")
            scale, scaler = self.discover_scale(self.data[col])
            axs[row_index, col_index].hist(self.data[col] / scaler, bins=50)
            axs[row_index, col_index].set_title(f"{clean_col} ({scale})")

        # Add a main title and adjust the subplot layout
        fig.suptitle(f"Distribution of Time Between Last (index=lag) Events ({time_col.title()})")
        fig.tight_layout()

        # Show the plot
        if plot:
            plt.show()
        plt.close()

    def plot_dates(self, time_col, time_freq, plot=True):
        # Get the time frequency and convert it to a pandas frequency string
        time_freq = time_freq[0].lower()
        if time_freq.endswith("s"):
            time_freq = time_freq[:-1]
        freq_str = \
            {"year": "%Y-01-01", "month": "%Y-%m-01", "week": "%Y-%U-0", "day": "%Y-%m-%d", "hour": "%Y-%m-%d %H",
             "minute": "%Y-%m-%d %H:%M", "second": "%Y-%m-%d %H:%M:%S"}[time_freq]

        # Set the figure size and style
        plt.figure(figsize=(9, 7))
        plt.style.use("seaborn-whitegrid")

        # Convert the time column to a pandas datetime column and format it to the requested time frequency
        column = pd.to_datetime(self.data[time_col]).dt.strftime(freq_str)

        # Count the number of occurrences of each date and sort the dates in ascending order
        counts = column.value_counts().sort_index()

        # Plot the data with appropriate title and x-axis label
        plt.plot(counts.index, counts.values, "o-")
        plt.title("Number of occurrences by {}".format(time_freq))
        plt.xlabel(time_freq.capitalize())
        plt.ylabel("Count")

        # Set the x-axis tick locations and labels to show a sample of the dates
        sample_size = min(20, len(counts))
        xticks = range(0, len(counts), len(counts) // sample_size)
        xtick_labels = [counts.index[i] for i in xticks]
        plt.xticks(xticks, xtick_labels, rotation=45)

        # Show the plot
        if plot:
            plt.show()
        plt.close()

    def plot_date_parts(self, time_col, plot=False):
        # Get the columns that represent date parts of the time column
        all_columns = self.data_info.get_all_features()
        optional_col_prefix = [
            'ts_dayinmonth_', 'ts_dayinweek_', 'ts_month_', 'ts_quarter_', 'ts_year', 'ts_minutes', 'ts_hours']
        date_part_cols = [name for name in all_columns if
                          any(name.startswith(prefix) for prefix in optional_col_prefix) and name.endswith(time_col)]

        # Create a subplot for each date part column and plot a heatmap of the value counts
        num_plots = len(date_part_cols)
        fig, axs = plt.subplots(num_plots, 1, figsize=(10, 4 * num_plots), sharey=False)
        for i, col in enumerate(date_part_cols):
            clean_col = remove_prefix_suffix(string=col, prefix="ts_", suffix=f"_{time_col}")

            # Count the number of occurrences of each unique value in the column
            counts = self.data[col].value_counts(normalize=True)[:14]

            # Get the unique values and their corresponding percentages
            values = counts.index

            # Sort the values in the order of the value labels
            values_sorted = sorted(values)

            # Reorder the corresponding value counts and percentages accordingly
            counts_sorted = counts.reindex(values_sorted)
            percentages_sorted = np.round(counts_sorted.values * 100).astype(int)

            # Create a matrix of percentages with one row and the same number of columns as unique values
            percent_matrix = np.array([percentages_sorted])

            # Plot the heatmap with appropriate title and axis labels
            sns.heatmap(percent_matrix, ax=axs[i], cmap="Blues", annot=True, fmt=".0f", cbar=False)

            # Set the x-axis tick labels to the unique values of the feature in the order of the value labels
            axs[i].set_xticklabels(values_sorted)

            axs[i].set_ylabel(clean_col)

        # Set the overall x-axis label and adjust the subplot layout
        fig.suptitle(f"Percentage of Occurrences by {time_col} Column Parts:")
        axs[-1].set_xlabel("Values")
        fig.tight_layout()

        # Show the plot
        if plot:
            plt.show()
        plt.close()

    def run_single(self, time_col, plot=True):
        self.plot_dates(time_col=time_col, time_freq=self.time_frequency, plot=plot)

    def run(self, plot=True):
        for time_col in self.time_cols:
            self.run_single(time_col, plot)
            self.plot_date_parts(time_col, plot)
            self.plot_time_between_events(time_col, plot)



class FeatureEliminationUtils:
    def __init__(self, df):
        self.df = df

    def eliminate_single_values_features(self):
        # Find the features with a single value
        single_value_features = []
        for col in self.df.columns:
            if self.df[col].nunique() == 1:
                single_value_features.append(col)

        # Report the features and remove them from the DataFrame
        if single_value_features:
            print(f"Features with a single value: {', '.join(single_value_features)}")
            self.df = self.df.drop(single_value_features, axis=1)

    def eliminate_high_shared_features(self, p=0.7):
        # Find the features with at least p percent of the same values
        high_correlation_features = []
        for col in self.df.columns:
            if self.df[col].value_counts(normalize=True).iloc[0] >= p:
                high_correlation_features.append(col)

        # Report the features and remove them from the DataFrame
        if high_correlation_features:
            print(f"Features with at least {p * 100}% of the same values: {', '.join(high_correlation_features)}")
            self.df = self.df.drop(high_correlation_features, axis=1)

    def eliminate_duplicates_features(self):
        # Find duplicated features
        duplicates = []
        for i, col1 in enumerate(self.df.columns):
            for col2 in self.df.columns[i + 1:]:
                if self.df[col1].equals(self.df[col2]):
                    duplicates.append(col2)

        # Report the duplicates and remove them from the DataFrame
        if duplicates:
            print(f"Duplicated features: {', '.join(duplicates)}")
            self.df = self.df.drop(duplicates, axis=1)

    def run(self):
        self.eliminate_single_values_features()
        self.eliminate_duplicates_features()
        self.eliminate_high_shared_features()
        return self.df

# import pandas as pd
# import numpy as np
#
# # Create a sample DataFrame with multiple features and duplicated columns
# data = {
#     'A': [1, 2, 3, 4, 5],
#     'B': [1, 1, 1, 1, 1],
#     'C': [3, 3, 3, 3, 3],
#     'D': [1, 1, 1, 1, 2],
#     'E': [np.nan, 2, 3, 4, 5],
#     'F': [1, 2, 3, 4, 5],
#     'G': [1, 2, 3, 4, 5],
#     'H': [1, 1, 1, 1, 1],
#     'I': [1, 2, 3, 4, 5],
#     'J': [np.nan, np.nan, np.nan, np.nan, np.nan],
# }
#
# df = pd.DataFrame(data)
#
# fe = FeatureEliminationUtils(df)
# print(fe.run())



import matplotlib.pyplot as plt
from wordcloud import WordCloud


# define a function that takes a dataframe and creates a word cloud from the 'text' column
def create_word_cloud(data, text_column='text'):
    # join the texts into a single string
    text = ' '.join(data[text_column])

    # create a WordCloud object
    wordcloud = WordCloud().generate(text)

    # plot the word cloud
    plt.figure(figsize=(14, 14))
    plt.imshow(wordcloud, interpolation='bilinear')
    plt.axis("off")
    plt.show()
    plt.close()



import math

import matplotlib.pyplot as plt
import pandas as pd

from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo
import seaborn as sns


class TimeSeriesPreModelAnalysis:
    def __init__(self, data_info: DataInfo, time_series_handler: TimeSeriesHandler):
        self.data_info = data_info
        self.time_series_handler = time_series_handler
        self.target_col = data_info.target
        self.target = self.data_info.data[self.target_col]
        self.date_cols = self.data_info.feature_types_val["time"]
        relevant_cols = [self.target_col] + \
                        list(self.data_info.feature_types_val["time"]) + \
                        [name for name in self.data_info.data.columns if name.startswith("ts_")]
        self.data = self.data_info.data[relevant_cols]

    def analyze_single_moving_metric(self, date_col):
        ts_cols = [name for name in self.data_info.data.columns if name.startswith("ts_")]
        moving_metric_str = "ts_mm_"
        _, metric_options = self.time_series_handler.lag_values

        relevant_cols = [name for name in ts_cols if
                         moving_metric_str in name and name.endswith(date_col) and metric_options[0] in name]
        print("analyze_single_moving_metric")
        df = self.data[[self.target_col] + list(set(relevant_cols))]
        if len(relevant_cols) > 0:
            self.plot_correlation_and_partial_dependence(df)
        plt.show()

    @staticmethod
    def plot_correlation_and_partial_dependence(df):
        # Separate the target column and the moving metric columns
        target_col = df.columns[0]
        moving_metric_cols = df.columns[1:]

        # Calculate the number of rows and columns for the partial dependence grid
        k = math.floor(math.sqrt(len(moving_metric_cols)))
        num_rows = num_cols = k

        # Create a figure with two subplots
        fig, axs = plt.subplots(nrows=2, figsize=(12, 8))

        # Plot the correlation heatmap
        corr_matrix = df.corr()[target_col][1:]
        sns.heatmap(pd.DataFrame(corr_matrix).transpose(), cmap="coolwarm", vmin=-1, vmax=1, annot=True, ax=axs[0])
        axs[0].set_title("Correlation between features and target")

        # Plot the partial dependence grid
        for i, moving_metric_col in enumerate(moving_metric_cols):
            if i >= k * k:
                plt.close()
                return None
            row = i // num_cols
            col = i % num_cols
            ax = fig.add_subplot(num_rows + 1, num_cols, num_cols + 1 + i)
            ax.plot(df[moving_metric_col], df[target_col], '.', alpha=0.1)
            ax.set_xlabel(moving_metric_col)
            ax.set_ylabel(target_col)
            if col != 0:
                ax.set_yticklabels([])
            if row != num_rows - 1:
                ax.set_xticklabels([])

        fig.suptitle("Correlation and partial dependence plots")
        fig.subplots_adjust(wspace=0.2, hspace=0.5)
        plt.show()
        plt.close()

    def analyze_moving_metric(self):
        for date_col in self.date_cols:
            self.analyze_single_moving_metric(date_col=date_col)

    def run(self):
        self.analyze_moving_metric()



import math

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from chester.zero_break.problem_specification import DataInfo


class CategoricalStats:
    def __init__(self, data_info: DataInfo, max_print=None):
        self.data_info = data_info
        self.max_print = max_print
        self.cols = list(set(self.data_info.feature_types_val["categorical"]))
        self.data = self.data_info.data[self.cols]
        if self.data.columns.duplicated().any():
            self.data = self.data.loc[:, ~ self.data.columns.duplicated()]
        self.data = self.data.sample(min(10000, len(self.data)))
        self.cols_sorted = self.sort_by_cardinality()

    def any_categorical(self):
        return True if len(self.cols) > 0 else False

    def sort_by_cardinality(self):
        if not self.any_categorical():
            return []
        cardinalities = []
        for col in self.cols:
            data = self.data[col]
            cardinalities.append((col, data.nunique()))
        sorted_cardinalities = sorted(cardinalities, key=lambda x: x[1], reverse=True)
        return [x[0] for x in sorted_cardinalities]

    def sample_top_features(self, n):
        top_features = self.cols_sorted[:3 * n]
        return top_features

    def plot_value_counts(self, n=25, plot=True):
        if not self.any_categorical():
            return None
        if not plot:
            return None
        top_n = self.cols_sorted[:min(len(self.cols_sorted), n)]
        num_plots = len(top_n)
        if num_plots == 1:
            col = top_n[0]
            fig, ax1 = plt.subplots(1, 1, figsize=(10, 5))
            cat_col = self.data[col].apply(lambda x: "cat " + str(x))
            count_data = cat_col.value_counts()
            percent_data = count_data / count_data.sum() * 100
            ax1.bar(count_data.index[:5], count_data.values[:5], color='gray')
            ax2 = ax1.twinx()
            ax2.plot(percent_data.index[:5], percent_data.values[:5], color='red', marker='o')
            ax1.set_ylabel('Counts', color='gray')
            ax2.set_ylabel('Percentages', color='red')
            ax1.set_xlabel(None)
            ax1.set_title(f"{col}")
            return None
        else:
            dim = max(math.ceil(math.sqrt(len(top_n))), 2)
            fig, ax = plt.subplots(dim, dim, figsize=(18, 15))
            fig.tight_layout()
            fig.suptitle("Top 5 Value Counts and Percentages for Each Feature")
            for i, col in enumerate(top_n):
                count_data = pd.DataFrame(self.data[col].value_counts()[0:5]).reset_index(drop=False)
                total_count = self.data[col].count()
                percent_data = count_data.copy()
                percent_data[col] = percent_data[col] / total_count * 100
                percent_data.rename(columns={col: 'percentage', 'index': col}, inplace=True)

                plot_title = f"{col}"
                ax_i = ax[i // dim, i % dim]
                ax1_i = ax_i
                ax1_i.bar(count_data.iloc[:, 0], count_data.iloc[:, 1].to_list(), color='gray')
                ax2_i = ax1_i.twinx()
                ax2_i.plot(percent_data.iloc[:, 0], percent_data.iloc[:, 1].to_list(), color='red', marker='o')
                ax1_i.set_ylabel('Counts', color='gray')
                ax2_i.set_ylabel('Percentages', color='red')
                ax1_i.set_xlabel(None)
                ax1_i.set_title(plot_title)
            # plt.tight_layout()
            plt.show()
            plt.close()

    def calculate_stats(self, is_print=True):
        from chester.util import ReportCollector, REPORT_PATH
        rc = ReportCollector(REPORT_PATH)
        if not self.any_categorical():
            return None
        result_dicts = []
        for col in self.cols:
            data = self.data[col]
            unique_values = data.nunique()
            missing_values = data.isnull().sum()
            value_counts = data.value_counts().rename("count").reset_index()
            value_counts["percentage"] = 100 * value_counts["count"] / value_counts["count"].sum()
            value_counts = value_counts.sort_values("count", ascending=False)
            dist_str = ', '.join([f"{row['index']}: {row['percentage']:.0f}%" for _, row in value_counts.iterrows()])
            result_dicts.append(
                {'col': col, '# unique': unique_values, '# missing': missing_values, 'Distribution': dist_str})
            data_unique_values = data.dropna().drop_duplicates()
            col_len = len(data_unique_values)
            values_to_sample = 3
            if col_len < values_to_sample:
                values_to_sample = col_len
            sample_values = [str(value) for value in data_unique_values.sample(min(col_len, values_to_sample)).values]
            result_dicts[-1]['Sample'] = ', '.join(sample_values)

            # add more columns
            # 1. % from all that covers the top 5 values
            top_5 = 100 * value_counts.iloc[:5]["count"].sum() / value_counts["count"].sum()
            result_dicts[-1][f'Top 5 values coverage'] = f"{top_5:.0f}%"

        results_df = pd.DataFrame(result_dicts)

        if is_print:
            if self.max_print is not None:
                formatted_df = format_df(df=results_df,
                                         max_value_width=self.max_print,
                                         )
                print(formatted_df)
            else:
                formatted_df = format_df(results_df)
                print(formatted_df)
            len_df = len(formatted_df)
            rc.save_object(obj=formatted_df.sample(min(len_df, 10)), text="Feature stats:")
        return results_df

    def run(self, plot=True):
        self.calculate_stats()
        self.plot_value_counts(plot=plot)
        return None


def format_df(df, max_value_width=30):
    pd.options.display.max_columns = None

    def trim_value(val):
        if len(str(val)) > max_value_width:
            return str(val)[:max_value_width] + "..."
        return str(val)

    df = df.applymap(trim_value)

    return df



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from pandas.errors import SettingWithCopyWarning

from chester.feature_stats.categorical_stats import CategoricalStats
from chester.feature_stats.numeric_stats import NumericStats
from chester.run.user_classes import TimeSeriesHandler
from chester.util import remove_prefix_suffix
from chester.zero_break.problem_specification import DataInfo


class TargetPreModelAnalysis:
    def __init__(self, data_info: DataInfo, time_series_handler: TimeSeriesHandler):
        import warnings
        warnings.filterwarnings("ignore", category=SettingWithCopyWarning)
        self.data_info = data_info
        self.time_series_handler = time_series_handler
        self.target_col = self.data_info.target
        self.target = self.data_info.data[self.target_col]
        relevant_cols = [self.target_col] + \
                        list(self.data_info.feature_types_val["time"]) + \
                        [name for name in self.data_info.data.columns if name.startswith("ts_")]
        relevant_cols = list(set(relevant_cols))
        print("Target relevant cols", relevant_cols)
        self.target_df = self.data_info.data[relevant_cols]  # select target, date and ts cols
        self.target_df.rename(columns={self.data_info.target: 'target_label'}, inplace=True)
        self.target_data_info = DataInfo(data=self.target_df.sample(min(len(self.target_df), 10000)))
        self.target_data_info.calculate()
        self.target_type = None
        # target type
        for feature_type, features in self.target_data_info.feature_types_val.items():
            if 'target_label' in features:
                self.target_type = feature_type

    def plot_histogram(self):
        target = self.target
        fig, ax = plt.subplots(figsize=(7, 7))
        ax.hist(target, bins=30, edgecolor='k', alpha=0.5, label='Histogram')
        ax.legend(loc='upper left')
        ax.set_xlabel('Values')
        ax.set_ylabel('Counts')
        ax.set_title(f'Histogram of {self.target.name}')
        ax2 = ax.twinx()
        sns.kdeplot(target, ax=ax2, label='Density')
        ax2.legend(loc='upper right')
        ax2.set_ylabel('Density')
        plt.show()
        plt.close()

    def plot_barplot(self):
        from chester.util import ReportCollector, REPORT_PATH
        rc = ReportCollector(REPORT_PATH)

        target = self.target
        value_counts = target.value_counts()

        # df_to_save = value_counts.to_dict(orient='records')[0]
        rc.save_object(obj=value_counts[0:10], text="values counts for categorical target to predict:")

        percentages = np.array(value_counts / target.size * 100)
        fig, ax1 = plt.subplots(figsize=(9, 9), dpi=100)
        ax2 = ax1.twinx()

        # Create a color map that shows the percentage of each target value
        cmap = sns.light_palette("green", as_cmap=True)
        heatmap = np.array([percentages, ] * len(value_counts))

        # Create a bar plot with a heatmap color scheme
        ax1.barh(value_counts.index, value_counts.values)
        ax1.set_xlabel('Counts')
        ax1.set_ylabel('Values')
        ax1.invert_yaxis()

        # Set the y-axis tick labels to the target values
        # ax1.set_yticklabels(value_counts.index)

        # Add a colorbar for the heatmap
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=percentages.min(), vmax=percentages.max()))
        sm._A = []
        cbar = fig.colorbar(sm, ax=ax1)
        cbar.ax.set_ylabel('Percentages')

        ax2.set_ylim([0, 100])
        ax2.plot(value_counts.index, percentages, color='red', marker='o')
        ax2.set_ylabel('Percentages', color='red')
        ax1.set_title(f'Bar Plot of {self.target.name}')
        plt.show()
        plt.close()

    def plot_dates(self, time_col, time_freq, plot=True):
        # Get the time frequency and convert it to a pandas frequency string
        time_freq = time_freq[0].lower()
        if time_freq.endswith("s"):
            time_freq = time_freq[:-1]
        freq_str = \
            {"year": "%Y-01-01", "month": "%Y-%m-01", "week": "%Y-%U-0", "day": "%Y-%m-%d", "hour": "%Y-%m-%d %H",
             "minute": "%Y-%m-%d %H:%M", "second": "%Y-%m-%d %H:%M:%S"}[time_freq]

        # Set the figure size and style
        plt.figure(figsize=(9, 7))
        plt.style.use("seaborn-whitegrid")

        # Convert the time column to a pandas datetime column and format it to the requested time frequency
        df = self.target_data_info.data.copy()
        df[time_col] = pd.to_datetime(df[time_col], format=freq_str)
        df = df.sort_values(time_col)

        # Count the number of occurrences of each date and sort the dates in ascending order
        counts = df[time_col].value_counts().sort_index()

        # Plot the data with appropriate title and x-axis label
        ax = plt.gca()
        ax.plot(counts.index, counts.values, "o-", label='Count')
        ax.set_xlabel(time_freq.capitalize())
        ax.set_ylabel("Count")

        # Set the x-axis tick locations and labels to show a sample of the dates
        sample_size = min(10, len(counts))
        xticks = range(0, len(counts), len(counts) // sample_size)
        xtick_labels = [counts.index[i].strftime(freq_str) for i in xticks]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xtick_labels, rotation=45)

        # Get the average of the target column and plot it on a secondary y-axis
        target_mean = self.target.groupby(df[time_col]).mean()
        ax2 = ax.twinx()
        ax2.plot(target_mean.index, target_mean.values, "o-", color="C1", label='Mean')
        ax2.set_ylabel(self.target_col, color="C1")
        ax2.tick_params(axis='y', labelcolor="C1")

        # Show the legend
        lines, labels = ax.get_legend_handles_labels()
        lines2, labels2 = ax2.get_legend_handles_labels()
        ax2.legend(lines + lines2, labels + labels2, loc='best')

        # Set the title
        ax.set_title("Number of occurrences by {} and mean {}".format(time_freq, self.target_col))

        # Show the plot
        if plot:
            plt.show()
        plt.close()

    def plot_date_parts(self, time_col, plot=False):
        # Get the columns that represent date parts of the time column
        df = self.target_data_info.data.copy()
        all_columns = self.data_info.get_all_features()
        optional_col_prefix = [
            'ts_dayinmonth_', 'ts_dayinweek_', 'ts_month_', 'ts_quarter_', 'ts_year', 'ts_minutes', 'ts_hours']
        date_part_cols = [name for name in all_columns if
                          any(name.startswith(prefix) for prefix in optional_col_prefix) and name.endswith(time_col)]

        # Create a subplot for each date part column and plot two rows: percentage of
        # occurrences and mean target variable
        num_plots = len(date_part_cols)
        fig, axs = plt.subplots(num_plots * 2, 1, figsize=(6, 5 * num_plots), sharey=False)

        for i, col in enumerate(date_part_cols):
            # Clean up the column name for plotting
            clean_col = remove_prefix_suffix(string=col, prefix="ts_", suffix=f"_{time_col}")

            # Get the unique values and their corresponding percentages of occurrences
            counts = df[col].value_counts(normalize=True)[:14]
            values = counts.index
            values_sorted = sorted(values)
            counts_sorted = counts.reindex(values_sorted)
            percentages_sorted = np.round(counts_sorted.values * 100).astype(int)
            percent_matrix = np.array([percentages_sorted])

            # Plot the heatmap of percentages with appropriate title and axis labels
            sns.heatmap(percent_matrix, ax=axs[i * 2], cmap="Blues", annot=True, fmt=".0f", cbar=False)
            axs[i * 2].set_xticklabels(values_sorted)
            axs[i * 2].set_ylabel(clean_col)

            # Get the average of the target variable for each date part
            target_mean = self.target.groupby(df[col]).mean()
            target_mean_sorted = target_mean.reindex(values_sorted)

            # Plot the target variable mean for each date part with appropriate title and axis labels
            axs[i * 2 + 1].plot(target_mean_sorted.index, target_mean_sorted.values, "o-", label='Mean')
            axs[i * 2 + 1].set_xticks(np.arange(len(values_sorted)))
            axs[i * 2 + 1].set_xticklabels(values_sorted)
            axs[i * 2 + 1].set_xlabel(clean_col)
            axs[i * 2 + 1].set_ylabel(self.target_col, color="C1")
            axs[i * 2 + 1].tick_params(axis='y', labelcolor="C1")

        # Set the overall x-axis label and adjust the subplot layout
        fig.suptitle(f"Percentage of Occurrences and Mean {self.target_col} by {time_col} Column Parts:")
        axs[-1].set_xlabel("Values")
        fig.tight_layout()

        # Show the plot
        if plot:
            plt.show()
        plt.close()

    def run(self, plot=True):
        if self.target_type == "numeric":
            print("Numerical statistics for target column")
            NumericStats(self.target_data_info).run(plot=False)
            if plot:
                self.plot_histogram()
                date_cols = self.data_info.feature_types_val["time"]
                if len(date_cols) > 0:
                    time_frequency = self.time_series_handler.time_frequency
                    for date_col in date_cols:
                        print("analyzing date_col", date_col)
                        self.plot_dates(time_col=date_col, time_freq=time_frequency, plot=plot)
                        self.plot_date_parts(time_col=date_col, plot=plot)

        elif self.target_type == "categorical":
            print("Categorical statistics for target column")
            CategoricalStats(self.target_data_info).run(plot=False)
            if plot:
                self.plot_barplot()



from typing import Tuple

import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.preprocessing import normalize


def get_bow_embedding(training_data: pd.DataFrame,
                      text_column: str = 'text',
                      test_data: pd.DataFrame = None,
                      ngram_range: Tuple[int, int] = (1, 1),
                      embedding_size=100):
    """
    Extract bag of words embeddings for the text in the given column of the DataFrame.

    Parameters
    ----------
    training_data: pd.DataFrame
        DataFrame containing the training data.
    test_data: pd.DataFrame, optional
        DataFrame containing the test data. If not provided, the function will only return the embeddings for the training data.
    text_column: str
        Name of the column in the DataFrame containing the text data.
    ngram_range: Tuple[int, int], optional
        Range of n-grams to consider. The default is (1, 1), which considers only unigrams.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame, 'Model']
        Tuple containing the bag of words embeddings for the train and test sets, and the trained CountVectorizer model.
    """
    # Initialize the CountVectorizer
    vectorizer = CountVectorizer(ngram_range=ngram_range, max_features=embedding_size)

    # Fit and transform the training data
    X_train = vectorizer.fit_transform(training_data[text_column])

    # Normalize the training data
    X_train = normalize(X_train)

    # Extract feature names
    feature_names = list(vectorizer.vocabulary_.keys())

    # Create a DataFrame for the training data
    train_embedding_df = pd.DataFrame(X_train.todense(), columns=[f"bow_{word}" for word in feature_names])

    if test_data is not None:
        # Transform the test data
        X_test = vectorizer.transform(test_data[text_column])

        # Normalize the test data
        X_test = normalize(X_test)

        # Create a DataFrame for the test data
        test_embedding_df = pd.DataFrame(X_test.todense(), columns=[f"bow_{word}" for word in feature_names])
        return train_embedding_df, test_embedding_df, vectorizer

    return train_embedding_df, None, vectorizer



from chester.preprocessing import preprocessing_func as pp
from chester.util import ReportCollector, REPORT_PATH
from chester.zero_break.problem_specification import DataInfo


class PreprocessHandler:
    def __init__(self, data_info: DataInfo, text_pre_process: pp.TextPreprocessor = None):
        self.data_info = data_info
        self.text_pre_process = text_pre_process
        if self.text_pre_process is not None:
            self.text_pre_process.df = self.data_info.data
        self.target = data_info.target
        self.problem_type_val = data_info.problem_type_val
        self.feature_types_val = data_info.feature_types_val
        self.loss_detector_val = data_info.loss_detector_val
        self.metrics_detector_val = data_info.metrics_detector_val
        self.model_selection_val = data_info.model_selection_val
        self.label_transformation_val = data_info.label_transformation_val

    def transform(self):
        text_columns = self.feature_types_val.get("text")
        for col in text_columns:
            if self.text_pre_process is not None:
                curr_text_pre_processor = self.text_pre_process
                curr_text_pre_processor.text_column = col
                text_reprocess = curr_text_pre_processor
            else:
                text_reprocess = pp.TextPreprocessor(self.data_info.data, text_column=col)
            title_to_print = f"{col} column preprocessing"
            print(title_to_print)
            rc = ReportCollector(REPORT_PATH)
            rc.save_text(title_to_print)
            text_reprocess.generate_report()
            self.data_info.data = pp.preprocess_text_df(text_reprocess)



import re


class FeatureTypeCategorizer:
    def __init__(self, feature_types: dict):
        self.feature_types = feature_types

    def categorize_features(self):
        new_features = {}
        for key, values in self.feature_types.items():
            category = self.categorize(key)
            if category in new_features:
                new_features[category].update(values)
            else:
                new_features[category] = set(values)
        self.feature_types = new_features
        return self.feature_types

    def categorize_numeric(self, input_string: str):
        input_string = input_string.lower()
        numeric_keywords = ['numeric', 'numerics', 'number', 'numbers', 'numerical', 'numericals', 'num']
        if any(re.search(r'\b' + keyword + r'\b', input_string, re.IGNORECASE)
               for keyword in numeric_keywords):
            return 'numeric'
        return None

    def categorize_boolean(self, input_string: str):
        input_string = input_string.lower()
        boolean_keywords = ['boolean', 'bool', 'booleany']
        if any(re.search(r'\b' + keyword + r'\b', input_string, re.IGNORECASE)
               for keyword in boolean_keywords):
            return 'boolean'
        return None

    def categorize_text(self, input_string: str):
        input_string = input_string.lower()
        text_keywords = ['text', 'nlp']
        if any(re.search(r'\b' + keyword + r'\b', input_string, re.IGNORECASE)
               for keyword in text_keywords):
            return 'text'
        return None

    def categorize_categorical(self, input_string: str):
        input_string = input_string.lower()
        categorical_keywords = ['categorical', 'categoric', 'cat', 'categorical', 'ordinal', 'ordinals']
        if any(re.search(r'\b' + keyword + r'\b', input_string, re.IGNORECASE)
               for keyword in categorical_keywords):
            return 'categorical'
        return None

    def categorize_time(self, input_string: str):
        input_string = input_string.lower()
        time_keywords = ['date', 'time', 'datetime', 'date time', 'time date', 'timestamp', 'time stamp',
                         'timestampped']
        if any(re.search(r'\b' + keyword + r'\b', input_string, re.IGNORECASE)
               for keyword in time_keywords):
            return 'time'
        return None

    def categorize(self, input_string: str):
        numeric = self.categorize_numeric(input_string)
        categorical = self.categorize_categorical(input_string)
        boolean = self.categorize_boolean(input_string)
        text = self.categorize_text(input_string)
        time = self.categorize_time(input_string)
        other = 'other'
        return numeric or categorical or boolean or text or time or other



import math

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.sparse as ss
from corextopic import corextopic as ct
from sklearn.feature_extraction.text import CountVectorizer
from wordcloud import WordCloud

from chester.util import ReportCollector, REPORT_PATH


def plot_corex_wordcloud(df, top_words=20, n_topics=10, plot=True, text_column='text'):
    top_words_list = get_top_words(df, top_words, n_topics, text_column=text_column)
    top_words_list = pd.DataFrame(top_words_list, columns=["topic_index", "term", "weight"])
    if plot:
        # Plot the word clouds
        N = math.floor(math.sqrt(n_topics))
        fig, axs = plt.subplots(N, N, figsize=(N * 4, N * 4), dpi=100)
        fig.suptitle("Word Clouds for Top Word for Corex Topics", fontsize=16)

        for i in range(n_topics - 1):
            topic_words = top_words_list[top_words_list["topic_index"] == i]
            topic_words = dict(zip(topic_words["term"], topic_words["weight"]))
            if len(topic_words) == 0:
                break
            wordcloud = WordCloud(width=800, height=800, background_color='black',
                                  stopwords=None, min_font_size=10).generate_from_frequencies(topic_words)

            subplot_index = i + 1
            try:
                ax = plt.subplot(N, N, subplot_index)
                plt.imshow(wordcloud)
                plt.axis("off")
                plt.title("Topic {}".format(i), fontsize=14)
            except:
                plt.close()
                pass
        plt.tight_layout()
        plt.show()
        plt.close()


def get_top_words(df: pd.DataFrame,
                  top_words,
                  n_topics,
                  max_features=1000,
                  text_column: str = 'text',
                  ngram_range=(1, 3)):
    rc = ReportCollector(REPORT_PATH)
    # Preprocess data
    vectorizer = CountVectorizer(stop_words='english',
                                 max_features=max_features,
                                 binary=True,
                                 ngram_range=ngram_range)
    doc_word = vectorizer.fit_transform(df[text_column])
    doc_word = ss.csr_matrix(doc_word)
    feature_names = list(vectorizer.vocabulary_.keys())
    words = list(np.asarray(feature_names))

    # Train model
    topic_model = ct.Corex(n_hidden=n_topics, words=words, max_iter=200, verbose=False, seed=1)
    topic_model.fit(doc_word, words=words)

    # Get top words and weights for each topic
    topics = topic_model.get_topics()
    if topics is None:
        return []

    top_words_list = []
    rc.save_text("\nCorex topic analysis:")
    for i, topic in enumerate(topics):
        if len(topic) == 0:
            break
        topic_words, weights, _ = zip(*topic)
        num_words = min(top_words, len(topic_words))  # Use smaller of n and num words in topic
        top_words_list += [(i, topic_words[j], weights[j]) for j in range(num_words)]
        title = '\tTopic {}: '.format(i + 1) + ', '.join(topic_words)
        print(title)
        rc.save_text(text=title)
    print("\n")

    return top_words_list
